/**
 * PETALS - PETALS Services Platform. Copyright (c) 2009 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id:
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.sftp.connection;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.mail.util.ByteArrayDataSource;

import org.ow2.petals.bc.sftp.SFTPClientPool;
import org.ow2.petals.bc.sftp.SftpClientPoolException;
import org.ow2.petals.bc.sftp.SshAuthenticationException;
import org.ow2.petals.component.framework.util.FileNamePatternUtil;
import org.ow2.petals.component.framework.util.FileNamePatternUtil.RegexFileNameFilter;

import com.sshtools.j2ssh.SftpClient;
import com.sshtools.j2ssh.sftp.SftpFile;

/**
 * Utility class that provides real implementation of the SFTP service.
 * 
 * 
 * @author Charles Casadei - eBM WebSourcing
 */
public class SFTPConnection {

    /**
     * A logger.
     */
    private Logger logger;

    /**
     * Information on the SFTP server.
     */
    private SFTPConnectionInfo info;

    /**
     * Default constructor.
     * 
     * @param info
     *            informations on the SFTP server
     * @param logger
     *            a logger
     * @throws java.io.IOException
     *             if any IO error occur during the process
     */
    public SFTPConnection(SFTPConnectionInfo info, Logger logger) throws IOException {
        super();
        this.info = info;
        this.logger = logger;
    }

    /**
     * Sends a set of files to a SFTP server.
     * 
     * @param attachments
     *            the file to be send
     * @return true if success
     * @throws SftpClientPoolException
     *             if a problem on the client pool occur while performing this
     *             operation
     * @throws SshAuthenticationException
     *             if the given authentication information are not valid
     * @throws IOException
     *             if any IO problem occur while sending file or connecting to
     *             server
     */
    public void mput(HashMap<String, DataHandler> attachments) throws SshAuthenticationException,
            SftpClientPoolException, IOException {

        // Process the attached files
        for (String fileName : attachments.keySet()) {
            logger.finest("Start sending : " + fileName + " file on the SFTP server.");

            // Make an InputStream from the attachment
            InputStream stream = attachments.get(fileName).getInputStream();

            // Sends the file to the server
            SftpClient client = SFTPClientPool.getSFTPClient(info, logger);
            synchronized (client) {
                try {
                    client.put(stream, fileName);
                } catch (IOException e) {
                    SFTPClientPool.cleanSFTPClient(info);
                    client = SFTPClientPool.getSFTPClient(info, logger);
                    synchronized (client) {
                        client.put(stream, fileName);
                    }
                }
            }
            logger.finest(fileName + " sent on the SFTP server");
        }
    }

    /**
     * Create a new file - using the given filename - on the SFTP server, with
     * the given content.
     * 
     * @param fileName
     *            name of the file to write onto the SFTP server
     * @param content
     *            file content, as String
     * @return true if everything seems to be OK
     * @throws SftpClientPoolException
     *             if a problem on the client pool occur while performing this
     *             operation
     * @throws SshAuthenticationException
     *             if the given authentication information are not valid
     * @throws IOException
     *             if any IO problem occur while sending file or connecting to
     *             server
     */
    public void putString(final String fileName, final String content)
            throws SshAuthenticationException, SftpClientPoolException, IOException {
        logger.finest("Start sending :" + fileName + " file on the SFTP server");

        // Makes an InputStream from the given content
        final InputStream stream = new ByteArrayInputStream(content.getBytes());

        SftpClient client = SFTPClientPool.getSFTPClient(info, logger);
        synchronized (client) {
            try {
                client.put(stream, fileName);
            } catch (IOException e) {
                client = SFTPClientPool.getSFTPClient(info, logger);
                synchronized (client) {
                    client.put(stream, fileName);
                }
            }
        }

        logger.finest(fileName + " sent on the FTP server");
    }

    /**
     * Retrieve the file given by its filename on the SFTP server.
     * 
     * @param remoteFile
     *            name of the file to get from the SFTP server
     * @return the retrieved file, streamed
     * @throws SftpClientPoolException
     *             if a problem on the client pool occur while performing this
     *             operation
     * @throws SshAuthenticationException
     *             if the given authentication information are not valid
     * @throws IOException
     *             if any IO problem occur while sending file or connecting to
     *             server
     */
    @SuppressWarnings("unchecked")
    public DataHandler get(final String remoteFile) throws SshAuthenticationException,
            SftpClientPoolException, IOException {

        // Holds the result of the GET operation
        DataHandler result = null;
        // Connect the client to the directory specified in the
        // SFTPConnectionInfo object
        SftpClient client = SFTPClientPool.getSFTPClient(info, logger);
        synchronized (client) {
            try {
                // Retrieve all the files from the current directory
                List<SftpFile> fileContainedWithinWorkingDir = client.ls();
                // Retrieve the files that match the given name
                for (SftpFile file : fileContainedWithinWorkingDir) {
                    if ((file.isFile() && file.canRead()) && file.getFilename().equals(remoteFile)) {
                        result = this.get(file, client);
                        break;
                    }
                }
            } catch (IOException e) {
                client = SFTPClientPool.getSFTPClient(info, logger);
                synchronized (client) {
                    // Retrieve all the files from the current directory
                    List<SftpFile> fileContainedWithinWorkingDir = client.ls();
                    // Retrieve the files that match the given name
                    for (SftpFile file : fileContainedWithinWorkingDir) {
                        if ((file.isFile() && file.canRead())
                                && file.getFilename().equals(remoteFile)) {
                            result = this.get(file, client);
                            break;
                        }
                    }

                }
            }
        }

        // If the file doen't exist
        if (result == null) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't get file : ");
            errorSB.append(remoteFile);
            errorSB.append(" from the SFTP server");
            throw new IOException(errorSB.toString());
        }

        return result;
    }

    /**
     * Retrieve the specified files from the SSH server.
     * 
     * @param fileFilterList
     *            the list of files to get from the SFTP server
     * @return fhe files specified in the list
     * @throws SftpClientPoolException
     *             if a problem on the client pool occur while performing this
     *             operation
     * @throws SshAuthenticationException
     *             if the given authentication information are not valid
     * @throws IOException
     *             if any IO problem occur while sending file or connecting to
     *             server
     */
    @SuppressWarnings("unchecked")
    public Map<String, DataHandler> mGet(final List<String> fileFilterList)
            throws SshAuthenticationException, IOException, SftpClientPoolException {

        // Holds the result of the MGET operation
        Map<String, DataHandler> dataHandlers = new HashMap<String, DataHandler>();

        // Performs the MGET operation

        // Construct the fileFilters to match files read from the FTP
        // directory
        List<FileNamePatternUtil.RegexFileNameFilter> filterList = new LinkedList<FileNamePatternUtil.RegexFileNameFilter>();
        for (String string : fileFilterList) {
            filterList.add(FileNamePatternUtil.getInstance()
                    .buildFileNameFilterFromWildChar(string));
        }

        RegexFileNameFilter filter = FileNamePatternUtil.getInstance()
                .buildFileNameFilterFromFilters(filterList);

        SftpClient client = SFTPClientPool.getSFTPClient(info, logger);
        synchronized (client) {
            try {
                List<SftpFile> filesContainedWithinWorkingDir = client.ls();
                for (SftpFile file : filesContainedWithinWorkingDir) {
                    if ((file.isFile() && file.canRead())
                            && filter.accept(null, file.getFilename())) {
                        dataHandlers.put(file.getFilename(), this.get(file, client));
                    }
                }
            } catch (IOException e) {
                client = SFTPClientPool.getSFTPClient(info, logger);
                List<SftpFile> filesContainedWithinWorkingDir = client.ls();
                for (SftpFile file : filesContainedWithinWorkingDir) {
                    if ((file.isFile() && file.canRead())
                            && filter.accept(null, file.getFilename())) {
                        dataHandlers.put(file.getFilename(), this.get(file, client));
                    }
                }
            }
        }

        return dataHandlers;
    }

    /**
     * List the content of the directory specified in the SFTPConfigurationInfo
     * used to build an instance of this object. Only the file are returned,
     * directories are ignored, because this is not relevant since this
     * component does not allow the user to swich from a directory to another.
     * 
     * @return the list of the files found in the "working directory" specified
     *         the SFTPConnectionInfo object which configures this connection
     * @throws SftpClientPoolException
     *             if a problem on the client pool occur while performing this
     *             operation
     * @throws SshAuthenticationException
     *             if the given authentication information are not valid
     * @throws IOException
     *             if any IO problem occur while sending file or connecting to
     *             server
     */
    @SuppressWarnings("unchecked")
    public List<String> list() throws SshAuthenticationException, SftpClientPoolException,
            IOException {

        // The list used to store the result of the ls operation
        List<String> fileList = new ArrayList<String>();

        // Connect the client to the directory specified in the
        // SFTPConnectionInfo object
        SftpClient client = SFTPClientPool.getSFTPClient(info, logger);

        List<SftpFile> sftpFiles;
        synchronized (client) {
            try {
                sftpFiles = client.ls();
            } catch (IOException e) {
                // retry
                client = SFTPClientPool.getSFTPClient(info, logger);
                sftpFiles = client.ls();
            }
        }

        for (SftpFile file : sftpFiles) {
            if (file.isFile()) {
                if (logger.isLoggable(Level.FINEST)) {
                    logger.finest("file : " + file.getFilename());
                }
                fileList.add(file.getFilename());
            }
        }

        return fileList;
    }

    /**
     * Streams the remote file.
     * 
     * @param remoteFile
     *            name of the remote file
     * @param client
     *            a SFTP client
     * @return a DataHandler that encapsulates the streamed remote file
     * @throws IOException
     *             if a problem occur while streaming / getting the remote file
     */
    private DataHandler get(SftpFile remoteFile, SftpClient client) throws IOException {
        DataHandler result;
        if (logger.isLoggable(Level.FINEST)) {
            logger.finest("Start receiving file : " + remoteFile.getFilename()
                    + " from SFTP server");
        }
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        client.get(remoteFile.getFilename(), outputStream);
        outputStream.flush();
        if (logger.isLoggable(Level.FINEST)) {
            logger.finest(remoteFile + " file received from SFTP server");
        }
        ByteArrayInputStream in = new ByteArrayInputStream(outputStream.toByteArray());
        DataSource dataSource = new ByteArrayDataSource(in, "application/octet-stream");
        result = new DataHandler(dataSource);
        if (logger.isLoggable(Level.FINEST)) {
            logger.finest(remoteFile + " set as DataSource");
        }
        return result;
    }

}
